/**
 * norris-tester
 * Copyright (c) 2011 Ruben L Z Tan
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a 
 * copy of this software and associated documentation files (the "Software"), 
 * to deal in the Software without restriction, including without limitation 
 * the rights to use, copy, modify, merge, publish, distribute, sublicense, 
 * and/or sell copies of the Software, and to permit persons to whom the 
 * Software is furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in 
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS 
 * IN THE SOFTWARE.
 */
'use strict';

var fs      = require('fs');
var path    = require('path');
var Emitter = require('events').EventEmitter;
var exec    = require('child_process').exec;
var vows    = require('vows');
var assert  = require('assert');
var mocker  = require(__dirname + '/../test/mock/base.mock');
var config  = JSON.parse(fs.readFileSync(__dirname + '/../config.json', 'utf-8'));

module.exports.make = function(deps) {

///////////////////////////////////////////////////////////////////////////////
//  Test runner module definition
///////////////////////////////////////////////////////////////////////////////
return {
    // Stores the options for this test runner
    options : null,

    //-------------------------------------------------------------------------
    // Configures the test runner
    config : function config(opt) {
        var self = this;

        // Sanity check on the options object
        if (undefined !== opt.input && -1 === config.input.indexOf(opt.input)) {
            throw new Error('Unsupported test framework (opt.input): ' + opt.input);
        }

        if (undefined !== opt.output && -1 === config.output.indexOf(opt.output)) {
            throw new Error('Unsupported reporting format (opt.output): ' + opt.output)
        }

        // Default set of options
        self.options = {
            dirs        : opt.dirs          || [],
            input       : opt.input         || 'vows',
            output      : opt.output        || 'console-pretty',
            report      : {
                passed      : opt.report.passed     || false,
                failed      : opt.report.failed     || false,
                skipped     : opt.report.skipped    || false,
                exception   : opt.report.exception  || false,
                verbose     : opt.report.verbose    || false
            }
        }

        return self;
    },

    //-------------------------------------------------------------------------
    // Runs the test runner
    run : function run(files, callback) {
        var self = this;

        if (null === self.options) {
            throw new Error('');
        }

        return self;
    }
}

/******************************************************************************
    Usage Examples:

    // Installation
    npm install norris-tester

    // Running it programmatically
    var tester = require('norris-tester').make();

    // Configure the options for the test runner
    tester.config({
        dirs : [
            './test/foo',
            './test/bar'
        ],
        testLib : 'vows',           // vows, nodeunit
        reportType : 'json',        // json, console, console-pretty, raw
        report : {
            passOnly    : false,
            failOnly    : false,
            exception   : false,
            verbose     : false
        }
    });

    // Collect the test cases that you want norris-tester to run
    var files = [
        './test/file1.js',
        './test/file2.js'
    ];

    // Run the tester
    tester.run(files, function(err, results) {
        // deal with the results
    });
 *****************************************************************************/
}










//=============================================================================
// Test module definition
var Test = module.exports = {
    // Properties
    types   : ['all', 'unit', 'hook', 'full'],
    options : ['p', 'pass-only', 'f', 'fail-only', 'v', 'verbose', 'e', 'exception', 'h', 'help'],
    
    // Methods
    init    : init,
    run     : run
};

//-----------------------------------------------------------------------------
// Initializes the command and options array to return an easier to handle
// JSON object
function init(util, commands, options) {
    var self    = this;
    var result  = {
        type        : commands[0],
        passOnly    : false,
        failOnly    : false,
        exception   : false,
        verbose     : false,
        help        : true,
        type        : 'all'
    };
    
    if (!config.type || -1 !== self.types.indexOf(type)) {
        // Check if the test that we want to run is a supported type or just a 
        // list of unit test files
        config.type = (null == config.type) ? 'all' : config.type;
    } else {
        // Not a supported type. Treat it as a namespace to a file 
        config.type = [];
        for (var i = commands.length; i--;) {
            var cmd = commands[i];
            config.type[cmd.replace(/\.g/g, '/') + '.test.js'] = true;
        }
    }
    
    // Check if there are unsupported options
    var unsupported = util.diff(options, self.options);
    if (unsupported.length > 0) {
        util.log('#red[Unsupported options ' + unsupported.join(', ') + ']\n', 'c');
    } else {
        // Options
        result.passOnly     = util.has(options, 'p', 'pass-only');
        result.failOnly     = util.has(options, 'f', 'fail-only');
        result.exception    = util.has(options, 'e', 'exception');
        result.verbose      = util.has(options, 'v', 'verbose');
        result.help         = util.has(options, 'h', 'help');
    }

    return result;
}

//-----------------------------------------------------------------------------
// Runs the command
function run(util, commands, options) {
    var self    = this;
    var config  = this.init(util, commands, options);
    
    if (config.help) {
        util.help('test');
        return;
    }
    
    var promise  = new Emitter();
    var reporter = {};
    
    (function initReporter(reporter) {
        var json         = {};
        reporter.results = [];
        reporter.reset   = function() {}
        reporter.report  = function(obj) { this.results.push(obj); }
        reporter.print   = function(str) {}
    })(reporter);
    
    // Handle errors when running the test cases
    promise.on('error', function onErr(err) {
        util.log('Unexpected error encountered');
        util.log(err.stack);
    });
    
    // Handle file event
    promise.on('file', function onFile(type, path, name) {
        var test = null;
            
        // Utility function to auto-append whitespaces
        function conformTo(m, size) {
            m = m + '';
            while (m.length < size) m = ' ' + m;
            return m;
        }
        
        function checkDone(util, reporter) {
            util.totalFiles--;
            if (util.totalFiles <= 0) {
                promise.emit('done', reporter.results);
            }
        }
        
        try {
            // Add isNotUndefined to assert
            assert.isNotUndefined = function isNotUndefined(actual, message) {
                if (actual === undefined) {
                    assert.fail(actual, undefined, message || "expected {actual} to be anything but {actual}", "===", assert.isNotUndefined);
                }
            }
            
            test = require(path).make(vows, assert, mocker.load(name));
            
            test.run({ reporter : reporter }, function onRun(result) {
                var pass    = conformTo(result.honored, 4);
                var fail    = conformTo(result.broken, 4);
                var err     = conformTo(result.errored, 4);
                var total   = conformTo(result.total, 4);
                var time    = result.time + 'ms';
                var color   = (result.broken > 0 || result.errored > 0) ? 'red' : 'cyan';
                
                util.log('#bold[#' + color + '[' + name + ']]', 'cn');
                
                // Calculate how many whitespaces to use
                var wsl     = 30 - name.length;
                var ws      = '';

                for (var i = wsl; i--;) ws += ' ';

                util.log(ws, 'cn');
                util.log('#green[Pass: ' + pass + ']\t#yellow[Fail: ' + fail + ']\t', 'cn');
                util.log('#red[Erro: ' + err + ']\t#blue[Total: ' + total + ']\t#magenta[Time: ' + time + ']', 'c');
                
                // Check if we've run all the test cases
                checkDone(util, reporter);
            });
        } catch(err) {
            checkDone(util, reporter);
        }
    });
    
    // Handle completion of test cases
    promise.on('done', function onDone(results) {
        if (verbose) {
            config.passOnly     = true;
            config.failOnly     = true;
            config.exception    = true;
        }
        
        if (config.passOnly || config.failOnly) {
            util.log('\nTest Results: ');
            verboseDetails(util, results, {
                pass        : config.passOnly,
                fail        : config.failOnly,
                exception   : config.exception,
                verbose     : config.verbose
            });
        }
    });
    
    // Run the test suites!
    if ('string' === typeof config.type) {
        promise = bulkRun(util, promise, config.type);
    } else {
        promise = selectiveRun(util, promise, config.type);
    }
}

//-----------------------------------------------------------------------------
// Runs all tests of a specific type
function bulkRun(util, promise, type) {
    if ('all' === type) {
        type = ['unit', 'hook', 'full'];
    } else {
        type = [type];
    }
        
    util.totalFiles = 0;
    
    function walk(basepath) {
        fs.readdir(basepath, function onReadDir(err, files) {
            if (err) {
                util.log('Attempted to load ' + basepath);
                promise.emit('error', new Error(err));
            }
            else {
                // Check if it's a directory or a file
                for (var i = 0, len = files.length; i < len; i++) {
                    var name        = files[i];
                    var fullPath    = basepath + '/' + name;
                    var stat        = fs.statSync(fullPath);

                    if (stat.isDirectory()) {
                        walk(fullPath);
                    } else {
                        util.totalFiles++;
                        promise.emit('file', 'bulk', fullPath, name);
                    }
                }
            }
        });
    }
    
    for (var i = type.length; i--;) {
        walk(path.normalize(__dirname + '/../test/' + type[i]));
    }
    
    return promise;
}

//-----------------------------------------------------------------------------
// Selectively run the test cases
function selectiveRun(util, promise, files) {
    var base = path.normalize(__dirname + '/../test');
    
    for (var i in files) {
        var file     = i;
        var fullPath = base + '/' + file;
        
        // Only process if the file exists
        if (path.existsSync(fullPath)) {
            // Get the filename
            var name = path.basename(file, path.extname(file));
            promise.emit('file', 'file', fullPath, name);
        } else {
            promise.emit('error', new Error(fullPath + ' not found'));
        }
    }
    
    return promise;
}

//-----------------------------------------------------------------------------
// Process the results of the test into something more usable
function processResults(results) {
    for (var i in results) {
        var result  = results[i];
        var type    = result.shift();
        
        if ('subject' === type) {
            suites[curr] = { title : result.shift(), status : true };
            continue;
        } else if ('finished' === type) {
            curr++;
            continue;
        }
                
        var data = result.shift();
        
        switch (type) {
            case 'context' :
                if (undefined === suites[curr].context) {
                    suites[curr].context = [];
                }
                suites[curr].context[data] = { title : data, status : true, vows : [] };
                break;
                
            case 'vow' :
                suites[curr].context[data.context].vows.push(data);
                if ('honored' !== data.status) {
                    var suite = suites[curr];
                    suite.status                        = false;
                    suite.context[data.context].status  = false;
                    suites[curr]                        = suite;
                }
                break;
            
            default : continue;
        }
    }

    return results;
}

//-----------------------------------------------------------------------------
// System output of the test results to integrate with chapel
function systemDetails(util, results, options) {
    
}

//-----------------------------------------------------------------------------
// Verbose output of the test results
function verboseDetails(util, results, options) {
    var suites  = [];
    var curr    = 0;
    
    for (var i in results) {
        var result  = results[i];
        var type    = result.shift();
        
        if ('subject' === type) {
            suites[curr] = { title : result.shift(), status : true };
            continue;
        } else if ('finished' === type) {
            curr++;
            continue;
        }
                
        var data = result.shift();
        
        switch (type) {
            case 'context' :
                if (undefined === suites[curr].context) {
                    suites[curr].context = [];
                }
                suites[curr].context[data] = { title : data, status : true, vows : [] };
                break;
                
            case 'vow' :
                suites[curr].context[data.context].vows.push(data);
                if ('honored' !== data.status) {
                    var suite = suites[curr];
                    suite.status                        = false;
                    suite.context[data.context].status  = false;
                    suites[curr]                        = suite;
                }
                break;
            
            default : continue;
        }
    }
    
    // Now that we have all the contexts buried in the right subjects, time to display them!
    for (var i in suites) {
        var suite   = suites[i];
        var title   = suite.title;
        var status  = suite.status;
        var color   = (status) ? 'cyan' : 'red';
        
        if (options.pass && status || options.fail && !status) {
            util.log('- #' + color + '[' + title + ']', 'c');
        }
            
        for (var ctxTitle in suite.context) {
            var context     = suite.context[ctxTitle];
            var ctxStatus   = context.status;
            var ctxColor    = (ctxStatus) ? 'magenta' : 'red';
            
            if (options.pass && ctxStatus || options.fail && !ctxStatus) {
                util.log('  • #' + ctxColor + '[' + ctxTitle + ']', 'c');
            }
                
            for (var j in context.vows) {
                var vow         = context.vows[j];
                var vowTitle    = vow.title;
                var vowStatus   = ('honored' === vow.status);
                var vowColor    = (vowStatus) ? 'green' : 'red';
                var vowIcon     = (vowStatus) ? '√' : '✗';
                var vowError    = vow.exception;
                
                if (options.pass && vowStatus || options.fail && !vowStatus) {
                    util.log('    #' + vowColor + '[' + vowIcon + ' ' + vowTitle + ']', 'c');
                }
                
                if (options.exception && !vowStatus && (!options.pass || options.verbose)) {
                    util.log('      #blue[' + vowError.replace(/\n/g, '\n      ') + ']', 'c');
                }
            }
        }
    }
    
}
